module m_isostasy
contains

subroutine isostasy (params,weightel,ov,surface,flag0,bcdef,isodef,istep,iter)

! first go at imposing isostasy
! this routine assumes that the original surface are flat
! it takes the zref height of each surface and the density of the corresponding material
!   to compute the reference isostatic column
! note also that this will not work when the courant condition is used
! flag0 is 1 if midpoint time step (dt/2) or 0 if full time step (dt)

use definitions
use m_octree
!use mpi

implicit none

include 'mpif.h'

type(parameters) params
type(octreev) ov
double precision weightel(ov%nleaves)
type(sheet) surface(params%ns)
integer flag0
type(bc_definition) bcdef
type (isostasy_definition) :: isodef
integer istep
integer iter

double precision eps,dtcur
integer i,j,ii,jj,kk,iproc,nproc,ierr
double precision,dimension(:),allocatable::diso
integer,dimension(:),allocatable::np
character shift*72
double precision,dimension(:,:),allocatable::weight,flex,work
double precision pi,hx,dh,pihx,xk,fi,fj,tij,d,isovelomax,isofrac
integer nflex
integer im,jm,ip,jp,nb
integer wisomaxtype
double precision :: wisomax, ledge, findx, findy
logical :: flat_edges

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

eps=1.d-8
shift=' '
nb=2**params%levelmax_oct

if (bcdef%bctype == 'iso_only') isovelomax = bcdef%bc_parameters(1)
if (isodef%isovelomaxtype > 9) then
    flat_edges = .true.
    ledge = 1.5d-3 * isodef%elastic_plate_thickness / params%length_scale
else
    flat_edges = .false.
end if

if (flag0==1) then
  dtcur=params%dt/2.d0
else
  dtcur=params%dt
endif

allocate (weight(nb,nb))

call calculate_weight (weight, params, ov, weightel)

if (.not.isodef%flexure) then
  ! apply local isostasy
  ! the deflection is stored in array weight
  do j=1,nb
    do i=1,nb
      if (isodef%local_isoref) then
        weight(i,j)=(weight(i,j)-isodef%weightref(i,j))/isodef%bottom_density
      else
        weight(i,j)=(weight(i,j)-isodef%weightref(1,1))/isodef%bottom_density
      end if
    enddo
  enddo
else
  ! flexural isostasy
  nflex=8*nb
  allocate(work(nflex,nflex),flex(nflex,nflex))
  hx=params%length_scale*8.d3
  xk=-isodef%bottom_density*params%density_scale*9.81d0
  d=1.d11/12.d0/(1.d0-0.25d0**2)*isodef%elastic_plate_thickness**3
  dh=hx/(nflex-1)
  
  !if (iproc.eq.0) print*,'hx,dh,xk,d',hx,dh,xk,d
  
  ii=nflex/2-nb/2-1
  jj=ii
  flex=0.d0
  do j=1,nb
    do i=1,nb
      if (isodef%local_isoref) then
        flex(ii+i,jj+j)=(weight(i,j)-isodef%weightref(i,j)) * dh*dh * &
                    params%density_scale*9.81d0*params%length_scale*1.d3
      else
        flex(ii+i,jj+j)=(weight(i,j)-isodef%weightref(1,1)) * dh*dh * &
                    params%density_scale*9.81d0*params%length_scale*1.d3
      end if
    enddo
  enddo
  
  !if (iproc.eq.0) print*,'poids',minval(flex),maxval(flex)
  
  do j=1,nflex
    call dsinft (flex(1,j),nflex)
  enddo
  
  do j=1,nflex
    do i=1,nflex
      work(j,i)=flex(i,j)
    enddo
  enddo
  
  do i=1,nflex
    call dsinft (work(1,i),nflex)
  enddo
  
  do j=1,nflex
    do i=1,nflex
      work(j,i)=work(j,i)*4.d0/hx/hx
    enddo
  enddo
  
  pi=atan(1.d0)*4.d0
  pihx=pi/hx
  do j=1,nflex
    fj=(j*pihx)**2
    do i=1,nflex
      fi=(i*pihx)**2
      tij=d/xk*(fi**2+2.d0*fi*fj+fj**2)+1.d0
      work(j,i)=work(j,i)/xk/tij
    enddo
  enddo
  
  do i=1,nflex
    call dsinft (work(1,i),nflex)
  enddo
  
  do j=1,nflex
    do i=1,nflex
      flex(i,j)=work(j,i)
    enddo
  enddo
  
  do j=1,nflex
    call dsinft (flex(1,j),nflex)
  enddo
  
  !if (iproc.eq.0) print*,'def in m',minval(flex),maxval(flex)
  
  ii=nflex/2-nb/2-1
  jj=ii
  do j=1,nb
    do i=1,nb
      weight(i,j)=-flex(ii+i,jj+j)/params%length_scale/1.d3
    enddo
  enddo
  
  deallocate (flex,work)
  
  !if (iproc.eq.0) print*,'dimensionless def ',minval(weight),maxval(weight)

endif

! redistributes the isostatic response onto the elements
! and distributes it onto the nodes into array diso(nnode)

allocate (np(ov%nnode))
allocate (diso(ov%nnode),work(nb+1,nb+1))

do j=1,nb+1
  do i=1,nb+1
    im=max(i-1,1)
    jm=max(j-1,1)
    ip=min(i,nb)
    jp=min(j,nb)
    work(i,j)=(weight(ip,jp)+weight(im,jp)+weight(ip,jm)+weight(im,jm))/4.d0
  enddo
enddo

do i=1,ov%nnode
  if (flat_edges) then
      findx = max(ledge, min(1.0d0-ledge, ov%x(i)+1.d-20))
      findy = max(ledge, min(1.0d0-ledge, ov%y(i)+1.d-20))
      call find_integer_coordinates(findx,findy,ov%z(i)+1.d-20,ii,jj,kk,params%levelmax_oct)
  else
      call find_integer_coordinates(ov%x(i)+1.d-20,ov%y(i)+1.d-20,ov%z(i)+1.d-20,ii,jj,kk,params%levelmax_oct)
  end if
  ii=ii+1
  jj=jj+1
  kk=kk+1
  if ((ii-1)*(ii-nb-1).gt.0 .or. (jj-1)*(jj-nb-1).gt.0) stop 'error in isostasy'
  diso(i)=work(ii,jj)
enddo


if (bcdef%bctype == 'iso_only' .and. istep==params%irestart+1 .and. iter==1) then
  if (maxval(diso) < (params%dt * isovelomax)) then
    params%nstep = 1
  else 
    params%nstep = int(maxval(diso)/(params%dt*isovelomax))+1
  endif
  write(*,'(a,i5,a)') shift//'Isostatic adjustment will require ',params%nstep,' time steps'
  write(*,'(a,f10.7)') shift//'Time step length is now ',params%dt
endif

if (bcdef%bctype == 'iso_only') then
  isofrac = 1.d0/real(params%nstep-(istep-1))
  work = work * isofrac
  diso = diso * isofrac
endif

! Store the total displacement of the basal node plane 
if (params%isobc) bcdef%zisoinc=work

! transforms the displacements into velocities and add this contribution to the velocity
! solution of the Stokes equation before interpolation onto the surfaces and their advection

wisomaxtype = mod(isodef%isovelomaxtype,10)
wisomax = maxval(abs(diso)/dtcur)
if (wisomaxtype == 2) then
    wisomax = maxval(abs(diso)/dtcur)
    if (wisomax < isodef%isovelomax) then
        wisomaxtype = 0
    end if
end if

if (wisomaxtype == 0) then
    do i=1,ov%nnode
      ov%wnodeiso(i)= diso(i)/dtcur
      ov%wnode(i)=ov%wnode(i)+ov%wnodeiso(i)
    enddo
else if (wisomaxtype == 1) then
    do i=1,ov%nnode
      ov%wnodeiso(i)= -sign(1.d0,diso(i)) * min( abs(diso(i)/dtcur), isodef%isovelomax)
      ov%wnode(i)=ov%wnode(i)+ov%wnodeiso(i)
    enddo
else if (wisomaxtype == 2) then
    do i=1,ov%nnode
      ov%wnodeiso(i)= isodef%isovelomax * COS(pi*diso(i) / (wisomax*dtcur)) 
      ov%wnode(i)=ov%wnode(i)+ov%wnodeiso(i)
    enddo
end if

if (params%debug>=1.and.iproc.eq.0) then
  write(*,'(a,f20.10,1x,f20.10)') shift//'Isostatic velocity range = ',minval(ov%wnodeiso),maxval(ov%wnodeiso)
  if(isodef%isovelomaxtype > 0) write(*,'(a,f20.10)') shift//'original iso-velocity range = ',wisomax
  write (*,*) 'displacement min: ',minval(ov%wnodeiso)*dtcur
  write (*,*) 'displacement max: ',maxval(ov%wnodeiso)*dtcur

  !output weightref and weights at end of timestep if desired
  if (flag0==0 .AND. params%debug>1) then
      call output_weights (weight, isodef%weightref, isodef%local_isoref, params)
  end if
endif

deallocate (np)
deallocate (diso)
deallocate (weight)
deallocate (work)

return
end subroutine isostasy

subroutine calculate_weight (weight, params, ov, weightel)

!------------------------------------------------------------------------------|
!((((((((((((((((  Purpose of the routine  )))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! Calculates the weight for each column in a nb x nb grid.
! nb = 2**levelmax_oct
!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
use definitions
use m_octree

implicit none
include 'mpif.h'

double precision, intent(out), dimension (:,:) :: weight
type(parameters), intent(in) :: params
type(octreev), intent(in) :: ov
double precision weightel(ov%nleaves)

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

integer :: i,ii, j,jj, ie, kk, nb
integer :: leaf,level,loc,dlev
integer :: nproc, iproc, ierr
double precision :: x0,y0,z0, xc,yc,zc, dd, dxyz

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

! first calculates the weight of each column, based on the weight of each elements
! this is performed on a 2**levelmax X 2**levelmax grid at the base of the model
! the result is stored in weight(i,j)

nb=2**params%levelmax_oct
dd=1.d0/2.d0**(params%levelmax_oct+1)
weight=0.d0

do ie=1,ov%nleaves
  xc=0.d0; yc=0.d0; zc=0.d0
  do i=1,8
    xc = xc + ov%x(ov%icon(i,ie))
    yc = yc + ov%y(ov%icon(i,ie))
    zc = zc + ov%z(ov%icon(i,ie))
  enddo
  xc=xc/8.d0
  yc=yc/8.d0
  zc=zc/8.d0
  call octree_find_leaf (ov%octree,ov%noctree,xc,yc,zc,leaf,level,loc,x0,y0,z0,dxyz)
  dlev=params%levelmax_oct-level
  call find_integer_coordinates (x0+dd,y0+dd,z0+dd,ii,jj,kk,params%levelmax_oct)
  ii=min(ii,nb-2**dlev)
  jj=min(jj,nb-2**dlev)
  do i=1,2**dlev
    do j=1,2**dlev
      weight(ii+i,jj+j)=weight(ii+i,jj+j)+weightel(leaf)/(4.d0**dlev)*(4.d0**params%levelmax_oct)
    enddo
  enddo
enddo

end subroutine calculate_weight


subroutine calculate_weightref (params, ov, weightel, isodef)

!------------------------------------------------------------------------------|
!((((((((((((((((  Purpose of the routine  )))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! Calculates the weightref for isostasy routine
!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
use definitions
use m_octree

implicit none
include 'mpif.h'

type(parameters), intent(in) :: params
type(octreev), intent(in) :: ov
double precision, intent(in) :: weightel(ov%nleaves)
type (isostasy_definition), intent(inout) :: isodef


!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

integer :: i, j, nb
integer :: nproc, iproc, ierr
integer, dimension(2) :: weightrefloc
double precision, allocatable, dimension (:,:) :: weight

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

nb=2**params%levelmax_oct

allocate (weight(nb,nb))

call calculate_weight (weight, params, ov, weightel)

if (isodef%local_isoref) then
    do j=1,nb
        do i=1,nb
            isodef%weightref(i,j)=weight(i,j)
        end do
    end do
else
    weightrefloc=minloc(abs(weight))
    isodef%weightref = weight(weightrefloc(1),weightrefloc(2))
endif

isodef%hasweightref = .true.

end subroutine calculate_weightref

subroutine output_weights (weight, weightref, local_isoref, params)

use definitions
implicit none

double precision, intent(in), dimension (:,:) :: weight, weightref
type(parameters) :: params
logical :: local_isoref

double precision :: dxy

integer :: nx, ny, i, j, ntot

character(len=4) cstep


nx = size(weight,DIM=1)
ny = size(weight,DIM=2)

dxy = params%length_scale * 2.0d0**(-params%levelmax_oct)


write (cstep,'(i0.4)') params%current_step

open(unit=30,file='DEBUG/FORCES/iso_weights_'//cstep//'.vtk')
write(30,'(a)') '# vtk DataFile Version 3.0'
write(30,'(a)') 'isostasy weights'
write(30,'(a)') 'ASCII'
write(30,'(a)') 'DATASET STRUCTURED_POINTS'
write(30,'(a,3i5)') 'DIMENSIONS ', nx, ny, 1
write(30,'(a,3ES13.6E2)') 'ORIGIN ', dxy*0.5d0, dxy*0.5d0, 0.d0
write(30,'(a,3ES13.6E2)') 'SPACING ', dxy, dxy, 1.d0
write(30,'(a)') ''
if (.not. local_isoref) then
     write(30,'(a)')'FIELD FieldData 1'
     write(30,'(a)')'weightref 1 1 double'
     write(30,'(ES13.6E2)') weightref(1,1)
     write(30,'(a)') ''
end if
write(30,'(a,i8)') 'POINT_DATA ', nx * ny

write(30,'(a)') 'SCALARS weight float 1'
write(30,'(a)') 'LOOKUP_TABLE default'
do j=1,ny
do i=1,nx
    write (30,'(ES13.6E2)') weight (i,j)
end do
end do

if (local_isoref) then
    write(30,'(a)') 'SCALARS weightref float 1'
    write(30,'(a)') 'LOOKUP_TABLE default'
    do j=1,ny
    do i=1,nx
        write (30,'(ES13.6E2)') weightref (i,j)
    end do
    end do
end if

close (30)



end subroutine output_weights 


! SINFT

! taken from numerical recipes (see book for further information)

! subroutines called:
! NONE

      SUBROUTINE DSINFT(Y,N)
      common /vocal/ ivocal
      REAL*8 WR,WI,WPR,WPI,WTEMP,THETA
      double precision Y(N)
      THETA=3.14159265358979D0/DBLE(N)
      WR=1.0D0
      WI=0.0D0
      WPR=-2.0D0*DSIN(0.5D0*THETA)**2
      WPI=DSIN(THETA)
      Y(1)=0.0
      M=N/2
      DO 11 J=1,M
        WTEMP=WR
        WR=WR*WPR-WI*WPI+WR
        WI=WI*WPR+WTEMP*WPI+WI
        Y1=WI*(Y(J+1)+Y(N-J+1))
        Y2=0.5*(Y(J+1)-Y(N-J+1))
        Y(J+1)=Y1+Y2
        Y(N-J+1)=Y1-Y2
11    CONTINUE
      CALL DREALFT(Y,M,+1)
      SUM=0.0
      Y(1)=0.5*Y(1)
      Y(2)=0.0
      DO 12 J=1,N-1,2
        SUM=SUM+Y(J)
        Y(J)=Y(J+1)
        Y(J+1)=SUM
12    CONTINUE
      RETURN
      END SUBROUTINE DSINFT

! REALFT

! taken from numerical recipes (see book for further information)

! subroutines called:
! NONE

      SUBROUTINE DREALFT(DATA,N,ISIGN)
      common /vocal/ ivocal
      REAL*8 WR,WI,WPR,WPI,WTEMP,THETA
      double precision DATA(*)
      THETA=6.28318530717959D0/2.0D0/DBLE(N)
      C1=0.5
      IF (ISIGN.EQ.1) THEN
        C2=-0.5
        CALL DFOUR1(DATA,N,+1)
      ELSE
        C2=0.5
        THETA=-THETA
      ENDIF
      WPR=-2.0D0*DSIN(0.5D0*THETA)**2
      WPI=DSIN(THETA)
      WR=1.0D0+WPR
      WI=WPI
      N2P3=2*N+3
      DO 11 I=2,N/2+1
        I1=2*I-1
        I2=I1+1
        I3=N2P3-I2
        I4=I3+1
        WRS=SNGL(WR)
        WIS=SNGL(WI)
        H1R=C1*(DATA(I1)+DATA(I3))
        H1I=C1*(DATA(I2)-DATA(I4))
        H2R=-C2*(DATA(I2)+DATA(I4))
        H2I=C2*(DATA(I1)-DATA(I3))
        DATA(I1)=H1R+WRS*H2R-WIS*H2I
        DATA(I2)=H1I+WRS*H2I+WIS*H2R
        DATA(I3)=H1R-WRS*H2R+WIS*H2I
        DATA(I4)=-H1I+WRS*H2I+WIS*H2R
        WTEMP=WR
        WR=WR*WPR-WI*WPI+WR
        WI=WI*WPR+WTEMP*WPI+WI
11    CONTINUE
      IF (ISIGN.EQ.1) THEN
        H1R=DATA(1)
        DATA(1)=H1R+DATA(2)
        DATA(2)=H1R-DATA(2)
      ELSE
        H1R=DATA(1)
        DATA(1)=C1*(H1R+DATA(2))
        DATA(2)=C1*(H1R-DATA(2))
        CALL DFOUR1(DATA,N,-1)
      ENDIF
      RETURN
      END SUBROUTINE DREALFT

! FOUR1

! this routine is taken directly out of numerical recipes
! see the book for further information

! subroutines called:
! NONE

      SUBROUTINE DFOUR1(DATA,NN,ISIGN)
      common /vocal/ ivocal
      REAL*8 WR,WI,WPR,WPI,WTEMP,THETA
      double precision DATA(*)
      N=2*NN
      J=1
      DO 11 I=1,N,2
        IF(J.GT.I)THEN
          TEMPR=DATA(J)
          TEMPI=DATA(J+1)
          DATA(J)=DATA(I)
          DATA(J+1)=DATA(I+1)
          DATA(I)=TEMPR
          DATA(I+1)=TEMPI
        ENDIF
        M=N/2
1       IF ((M.GE.2).AND.(J.GT.M)) THEN
          J=J-M
          M=M/2
        GO TO 1
        ENDIF
        J=J+M
11    CONTINUE
      MMAX=2
2     IF (N.GT.MMAX) THEN
        ISTEP=2*MMAX
        THETA=6.28318530717959D0/(ISIGN*MMAX)
        WPR=-2.D0*DSIN(0.5D0*THETA)**2
        WPI=DSIN(THETA)
        WR=1.D0
        WI=0.D0
        DO 13 M=1,MMAX,2
          DO 12 I=M,N,ISTEP
            J=I+MMAX
            TEMPR=SNGL(WR)*DATA(J)-SNGL(WI)*DATA(J+1)
            TEMPI=SNGL(WR)*DATA(J+1)+SNGL(WI)*DATA(J)
            DATA(J)=DATA(I)-TEMPR
            DATA(J+1)=DATA(I+1)-TEMPI
            DATA(I)=DATA(I)+TEMPR
            DATA(I+1)=DATA(I+1)+TEMPI
12        CONTINUE
          WTEMP=WR
          WR=WR*WPR-WI*WPI+WR
          WI=WI*WPR+WTEMP*WPI+WI
13      CONTINUE
        MMAX=ISTEP
      GO TO 2
      ENDIF
      RETURN
      END SUBROUTINE DFOUR1

end module m_isostasy

