module m_move_surface
contains
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         |
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              MOVE_SURFACE    Nov. 2007                                       |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

subroutine move_surface (surface,surface0,flag0,flag1,ov,dt,params,istep,is)
!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! subroutine to move a set of particles on a surface from a velocity
! field known at the nodes of an octree
!
! surface is the surface on which the particles need to be moved
! surface0 is the surface in its initial geometry
! flag0 is 1 if in midpoint configuration and 0 if at the end of the timestep
! if flag0=0, then the surface0 arrays are deallocated!!
! flag1 is 1 if surface is only being shifted vertically due to isostasy
! octree is the velocity octree
! noctree is its size
! unode,vnode and wnode are the components of the velocity
! wnodeiso is the isostasy velocity (vertical only)
! wnodecompact is the compaction velocity (vertical only)
! nnode is the number of nodes on the octree
! icon_octree is the connectivity matrix on the octree
! nleaves is the number of leaves in the octree
! dt is the time step
! niter_move is the number of iterations used to move the particles
! it is read in input.txt
!
! here we use a simple method that performs a given (niter_move)
! number of iterations in a mid_point algorithm
!
! we also advect the normals by using the velocity gradient computed
! from the finite element shape function derivatives
!
!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|

use definitions
use m_toolbox
use m_compute_normals
use m_remove_multiple_points
use m_check_delaunay
use m_DoRuRe
use m_octree
use m_c_cloud
use m_create_surfaces
use m_remove_bad

!use mpi

implicit none

include 'mpif.h'

type (sheet), dimension(:) ::  surface
type (sheet), dimension(:) :: surface0
type (octreev) ov
integer flag0,flag1
double precision dt
type(parameters) params
integer istep,is
!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

integer i,iter,nedge,nrem,counter
integer err,nside,ref_count,nremoved
integer iproc,nproc,ierr
integer convexhull
integer,dimension(:),allocatable :: remove_list
double precision u,v,w,rat,eps
double precision x0,y0,z0,xi,yi,zi,x,y,z
double precision,dimension(:),allocatable::xbuf,ybuf,zbuf
double precision,dimension(:),allocatable::xnbuf,ynbuf,znbuf
double precision,dimension(:),allocatable::remaddlist
double precision ux,uy,uz,vx,vy,vz,wx,wy,wz,xyzn
double precision xn0,yn0,zn0,xyn,theta,phi
double precision xn1,yn1,zn1,xn2,yn2,zn2
double precision xln1,yln1,zln1,xln2,yln2,zln2
logical :: normaladvect
logical,dimension(:),allocatable :: remove
character*72  :: shift

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

!disable normaladvect for stacking surfaces and C Fast Advection
normaladvect = (params%normaladvect .AND. &
    (.NOT. (surface(is)%stacking .AND. surface(is)%nstack > 0)) .AND. &
    (params%particle_advect_method < 2))

!Fastscape surfaces may only move vertically and must not be refined
if (surface(is)%fastscape .AND. flag1 == 0) then
    write (*,*) 'ERROR: Fastscape Surface in move_surfaces'
    write (8,*) 'ERROR: Fastscape Surface in move_surfaces'
    return
end if

shift=' '

rat=1.d0-1.d-10
eps=1.d-10

!nside=2**surface(is)%levelt+1
nside=2**surface(is)%levelt+2  !opla

if (params%particle_advect_method > 1) then
    allocate (zbuf(surface(is)%nsurface),stat=err) 
    zbuf = surface(is)%z
    if (err.ne.0) call stop_run ('Error alloc zbuf in move_surface')

    if (flag1==0) then
        call c_move_surface (flag1, surface(is)%fastscape, surface(is)%convexhull, &
            surface(is)%nsurface, surface(is)%x, surface(is)%y, surface(is)%z, &
            ov%icon, ov%octree, ov%x, ov%y, ov%z, ov%unode, ov%vnode, ov%wnode, &
            params%vex, dt, normaladvect)
    elseif (flag1==1) then
        ! Only vertical velocity from isostasy
        call c_move_surface (flag1, surface(is)%fastscape, surface(is)%convexhull, &
            surface(is)%nsurface, surface(is)%x, surface(is)%y, surface(is)%z, &
            ov%icon, ov%octree, ov%x, ov%y, ov%z, ov%unode, ov%vnode, ov%wnodeiso, &
            params%vex, dt, normaladvect)
        !no isostasy motion for bottom particles
        if (surface(is)%bottom > 1) then
            where (zbuf < eps) surface(is)%z = 0.d0
        end if
    elseif (flag1==2) then
        ! Only vertical velocity from compaction
        call c_move_surface (flag1, surface(is)%fastscape, surface(is)%convexhull, &
            surface(is)%nsurface, surface(is)%x, surface(is)%y, surface(is)%z, &
            ov%icon, ov%octree, ov%x, ov%y, ov%z, ov%unode, ov%vnode, ov%wnodecompact, &
            params%vex, dt, normaladvect)
    else
        ! We have a problem if we get here...
        call stop_run ('Error: Unsupported value for flag1 in move_surface')
    endif

    !move or reset particles based on additional properties

    do i=1,surface(is)%nsurface
        !restrain motion for convex hull, corners are true for both
        if (surface(is)%closed == 1) then
            !Closed surfaces have no convex hull and will stay inside model domain
            surface(is)%x(i) = max(0.d0, min(1.d0, surface(is)%x(i)))
            surface(is)%y(i) = max(0.d0, min(1.d0, surface(is)%y(i)))
            surface(is)%z(i) = max(0.d0, min(1.d0, surface(is)%z(i)))
        end if

        !Bottom surfaces are only constained for z=0 plane
        if (surface(is)%bottom > 0) then
            surface(is)%z(i) = max(0.d0, surface(is)%z(i))
        end if
    end do

    deallocate (zbuf)
else
    allocate (xbuf(surface(is)%nsurface),stat=err)
    if (err.ne.0) call stop_run ('Error alloc xbuf in move_surface')
    allocate (ybuf(surface(is)%nsurface),stat=err)
    if (err.ne.0) call stop_run ('Error alloc ybuf in move_surface')
    allocate (zbuf(surface(is)%nsurface),stat=err)
    if (err.ne.0) call stop_run ('Error alloc zbuf in move_surface')
    allocate (xnbuf(surface(is)%nsurface),stat=err)
    if (err.ne.0) call stop_run ('Error alloc xnbuf in move_surface')
    allocate (ynbuf(surface(is)%nsurface),stat=err)
    if (err.ne.0) call stop_run ('Error alloc ynbuf in move_surface')
    allocate (znbuf(surface(is)%nsurface),stat=err)
    if (err.ne.0) call stop_run ('Error alloc znbuf in move_surface')
    
    xbuf=0.d0
    ybuf=0.d0
    zbuf=0.d0
    xnbuf=0.d0
    ynbuf=0.d0
    znbuf=0.d0
    xbuf=0.d0
    ybuf=0.d0
    zbuf=0.d0


    do i=iproc+1,surface(is)%nsurface,nproc
       u=0.d0
       v=0.d0
       w=0.d0
       x0=surface(is)%x(i)
       y0=surface(is)%y(i)
       z0=surface(is)%z(i)
       convexhull=surface(is)%convexhull(i)
       x=x0
       y=y0
       z=z0
       do iter=1,params%niter_move
          xi=(x0+x)/2.d0
          yi=(y0+y)/2.d0
          zi=(z0+z)/2.d0
          if (flag1==0) then
            ! Full velocity field
            call octree_interpolate (ov%octree,ov%icon, xi,yi,zi, &
                                          ov%unode,u, &
                                          ov%vnode,v, &
                                          ov%wnode/params%vex,w)
          elseif (flag1==1) then
            ! Only vertical velocity from isostasy
            call octree_interpolate (ov%octree,ov%icon, xi,yi,zi, &
                                          0.d0*ov%unode,u, &
                                          0.d0*ov%vnode,v, &
                                          ov%wnodeiso/params%vex,w)

            !no isostasy motion for bottom particles
            if (surface(is)%bottom > 1 .AND. zi < eps) w = 0.d0
          elseif (flag1==2) then
            ! Only vertical velocity from compaction
            call octree_interpolate (ov%octree,ov%icon, xi,yi,zi, &
                                          0.d0*ov%unode,u, &
                                          0.d0*ov%vnode,v, &
                                          ov%wnodecompact/params%vex,w)
          else
            ! We have a problem if we get here...
            call stop_run ('Error: Unsupported value for flag1 in move_surface')
          endif

          x=x0+dt*u
          y=y0+dt*v
          z=z0+dt*w

          !restrain motion for convex hull, corners are true for both
          if (btest(convexhull,1) .OR. btest(convexhull,3)) then  !y=0 or y=1
              y = y0
          end if
          if (btest(convexhull,2) .OR. btest(convexhull,4)) then  !x=0 or x=1
              x = x0
          end if


          !Closed surfaces have no convex hull and will stay inside model domain
          if (surface(is)%closed == 1) then
              x = max(0.d0, min(1.d0, x))
              y = max(0.d0, min(1.d0, y))
              z = max(0.d0, min(1.d0, z))
          end if

          !Bottom surfaces are only constained for z=0 plane
          if (surface(is)%bottom > 0) then
              z = max(0.d0, z)
          end if

       enddo
       xbuf(i)=x
       ybuf(i)=y
       zbuf(i)=z
    enddo

    if (normaladvect) then
       do i=iproc+1,surface(is)%nsurface,nproc
          if (flag1==0) then
            ! Full velocity field
            call octree_interpolate_derivative (ov%octree,ov%icon, xi,yi,zi, &
                                                     ov%unode,u,ux,uy,uz, &
                                                     ov%vnode,v,vx,vy,vz, &
                                                     ov%wnode/params%vex,w,wx,wy,wz)
          elseif (flag1==1) then
            ! Only vertical velocity from isostasy
            call octree_interpolate_derivative (ov%octree,ov%icon, xi,yi,zi, &
                                                     0.d0*ov%unode,u,ux,uy,uz, &
                                                     0.d0*ov%vnode,v,vx,vy,vz, &
                                                     ov%wnodeiso/params%vex,w,wx,wy,wz)
          elseif (flag1==2) then
            ! Only vertical velocity from compaction
            call octree_interpolate_derivative (ov%octree,ov%icon, xi,yi,zi, &
                                                     0.d0*ov%unode,u,ux,uy,uz, &
                                                     0.d0*ov%vnode,v,vx,vy,vz, &
                                                     ov%wnodecompact/params%vex,w,wx,wy,wz)
          else
            ! We have a problem if we get here...
            call stop_run ('Error: Unsupported value for flag1 in move_surface')
          endif

          xn0=surface(is)%xn(i)
          yn0=surface(is)%yn(i)
          zn0=surface(is)%zn(i)
          xyn=xn0**2+yn0**2
          if (xyn.gt.0.d0) xyn=sqrt(xyn)
          theta=atan2(xyn,zn0)
          phi=0.d0
          if (sin(theta).ne.0.d0) phi=atan2(yn0,xn0)
          xn1=cos(theta)*cos(phi)
          yn1=cos(theta)*sin(phi)
          zn1=-sin(theta)
          xn2=-sin(phi)
          yn2=cos(phi)
          zn2=0.d0
          xln1=ux*xn1+uy*yn1+uz*zn1
          yln1=vx*xn1+vy*yn1+vz*zn1
          zln1=wx*xn1+wy*yn1+wz*zn1
          xln2=ux*xn2+uy*yn2+uz*zn2
          yln2=vx*xn2+vy*yn2+vz*zn2
          zln2=wx*xn2+wy*yn2+wz*zn2
          xnbuf(i)=xn0+(yn1*zln2-zn1*yln2+yln1*zn2-zln1*yn2)*dt+(yln1*zln2-zln1*yln2)*dt**2
          ynbuf(i)=yn0+(zn1*xln2-xn1*zln2+zln1*xn2-xln1*zn2)*dt+(zln1*xln2-xln1*zln2)*dt**2
          znbuf(i)=zn0+(xn1*yln2-yn1*xln2+xln1*yn2-yln1*xn2)*dt+(xln1*yln2-yln1*xln2)*dt**2
          if (xnbuf(i).gt.1.d50) write(*,*) 'on proc ', iproc, ', xnbuf is large: ', xnbuf(i)
          if (ynbuf(i).gt.1.d50) write(*,*) 'on proc ', iproc, ', ynbuf is large: ', ynbuf(i)
          if (znbuf(i).gt.1.d50) write(*,*) 'on proc ', iproc, ', znbuf is large: ', znbuf(i)
          xyzn=sqrt(xnbuf(i)**2+ynbuf(i)**2+znbuf(i)**2)
          xnbuf(i)=xnbuf(i)/xyzn
          ynbuf(i)=ynbuf(i)/xyzn
          znbuf(i)=znbuf(i)/xyzn

          if (surface(is)%bottom > 0 .AND. z<=eps) then
              xnbuf(i) = 0.d0
              ynbuf(i) = 0.d0
              znbuf(i) = 1.d0
          end if

          !Since bottom points are not advected normally, the normals also have to be
          !set, best to the z=0-plane normal
          if (surface(is)%bottom > 0 .AND. z<=eps) then
              xnbuf(i) = 0.d0
              ynbuf(i) = 0.d0
              znbuf(i) = 1.d0
          end if
       enddo
    endif

    surface(is)%x=0.d0
    surface(is)%y=0.d0
    surface(is)%z=0.d0
    call mpi_allreduce (xbuf,surface(is)%x,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce (ybuf,surface(is)%y,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce (zbuf,surface(is)%z,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

    if (normaladvect) then
       surface(is)%xn=0.d0
       surface(is)%yn=0.d0
       surface(is)%zn=0.d0
       call mpi_allreduce (xnbuf,surface(is)%xn,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
       call mpi_allreduce (ynbuf,surface(is)%yn,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
       call mpi_allreduce (znbuf,surface(is)%zn,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    end if
    !else calculate normals at end of fix_surface_after_move

    if (normaladvect) then
       surface(is)%xn=0.d0
       surface(is)%yn=0.d0
       surface(is)%zn=0.d0
       call mpi_allreduce (xnbuf,surface(is)%xn,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
       call mpi_allreduce (ynbuf,surface(is)%yn,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
       call mpi_allreduce (znbuf,surface(is)%zn,surface(is)%nsurface,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    end if
    !else calculate normals at end of fix_surface_after_move

    deallocate (xbuf)
    deallocate (ybuf)
    deallocate (zbuf)
    deallocate (xnbuf)
    deallocate (ynbuf)
    deallocate (znbuf)
end if

!if stacking surface, now apply stacking again
if (surface(is)%stacking .AND. surface(is)%nstack > 0) then
  !stack surface with surface below
  if (iproc == 0) write (*,*) 'Adjust Surface', is, 'with underlying surface'
  surface(is)%z = surface(is)%z * params%vex
  call zpoints(surface,is,params%ns, surface(is)%nstack)
  surface(is)%z = surface(is)%z / params%vex
end if

!-------------------------------------------------------------------------------
! We now check whether points have been advected outside the simulation domain
! (unit cube), and if so, we need to remove them. However, because the triangulation
! is only computed at startup and further updated, the removal of a point requires
! the triangulation to be sown back by hand. This will most probably lead to a
! non-Delaunay triangulation (locally), but this will be taken care of in the
! check_delaunay subroutine.
! The corner nodes the convex hull are exempt from this

if (surface(is)%closed.eq.0) then
    !distable remove and refine for FastScape erosion surface
    if (.NOT. surface(is)%fastscape) then
        allocate (remove(surface(is)%nsurface))
        remove = .false.

        !find all nodes outside of the domain
        do i=1,surface(is)%nsurface
            !exclude corner nodes
            if (btest(surface(is)%convexhull(i), 0)) cycle
            if (surface(is)%x(i).lt.-eps .or. surface(is)%x(i).gt.1.d0+eps .or. &
              surface(is)%y(i ).lt.-eps .or. surface(is)%y(i).gt.1.d0+eps .or. &
              surface(is)%z(i).lt.-eps .or. surface(is)%z(i).gt.1.d0+eps) then
                remove(i)=.true.
            endif
        end do

        nrem=count(remove)
        if (nrem/=0) then
            !create remove_list
            allocate (remove_list(nrem))
            remove_list=0

            counter=0
            do i=1,surface(is)%nsurface
                if (remove(i)) then
                    counter = counter+1
                    remove_list(counter) = i
                end if
            end do

            if (params%debug > 1) then
                allocate (remaddlist(surface(is)%nsurface))
                remaddlist = 0.d0

                do i=1,nrem
                    remaddlist(remove_list(i)) = -1.d0
                end do

                call output_surf_nodevar(surface(is),is,'move',istep,1,remaddlist)

                deallocate (remaddlist)
            end if

            if (iproc == 0) write(*,'(a,i4,a)') shift//'removing',nrem,' particle from surface'

            call remove_multiple_points (remove_list, nrem, surface(is), flag0, surface0(is), is, params, 0)

            if (params%debug > 1) then
                call output_surf(surface(is),is,'move',istep,2)
            end if

            deallocate(remove_list)
        end if

        deallocate(remove)
    else
        nrem = 0
    end if

    if (surface(is)%checkdelaunay) then
        ref_count = 1
        nremoved = 0
        do
            call check_delaunay (params,surface(is),is,istep,'delaunay_after_move',ref_count)
            ! if no remove_bad_triangles was necessary, continue normally. Otherwise, redo refine_surface.
            if (ref_count > 0) then
                exit
            else
                call remove_bad_triangles (surface(is), surface0(is), flag0, is, params, 1, nremoved)
            end if
        end do
        nrem = nrem + nremoved
    end if

    if (.not. normaladvect) then
        call compute_normals (surface(is)%nsurface,surface(is)%x,surface(is)%y,surface(is)%z, &
            surface(is)%nt,surface(is)%icon,surface(is)%xn,surface(is)%yn,surface(is)%zn)
    end if

    if (flag0==0) then
        nedge=surface(is)%nsurface+surface(is)%nt-1
        call DoRuRe_surf_stats (params%doDoRuRe,istep,50,is,surface(is)%nt,surface(is)%nsurface,nedge,-nrem)
    end if

    if (iproc.eq.0) then
        write (8,*) 'Removal after move complete'
    end if
endif

end subroutine move_surface


end module m_move_surface
